import logging
import re
from typing import List, Optional, Pattern

from .loaders import Loader, Load


logger = logging.getLogger(__name__)


try:
    SecretManagerServiceClient = __import__(
        'google.cloud.secretmanager.SecretManagerServiceClient'
    )
    NotFound = __import__('google.api_core.exceptions.NotFound')
except ImportError:
    SecretManagerServiceClient = None
    NotFound = None


def get_google_cloud_secret(
    secret_id: str,
    google_project: str,
) -> Optional[str]:
    if SecretManagerServiceClient and NotFound:
        try:
            client = SecretManagerServiceClient()
            name = client.secret_version_path(
                google_project,
                secret_id,
                'latest',
            )
            response = client.access_secret_version(request={'name': name})
            secret = response.payload.data.decode('UTF-8')
            return secret
        except NotFound:
            logger.warning(
                f'Secret {secret_id} not found in project {google_project}'
            )
        return None
    logger.error(
        'google-cloud-secret-manager is not installed! '
        "Try to use 'pip install \"dyacon[google]\"' "
        "or 'pip install google-cloud-secret-manager' for installation"
    )
    return None


GOOGLE_LOADER_PATTERN = re.compile(
    r'!{'
    r'(?P<name>[^}^{:]+)'
    r'(?P<first_separator>:?)'
    r'(?P<project_id>[^}^{:]+)'
    r'(?P<second_separator>:?)'
    r'(?P<default>.*?)'
    r'}',
)


class GoogleLoader(Loader):
    def __init__(self) -> None:
        self._pattern = GOOGLE_LOADER_PATTERN

    @property
    def pattern(self) -> Pattern[str]:
        return self._pattern

    def load(self, config_content: str) -> str:
        findings = list(self.pattern.finditer(config_content))

        for match in reversed(findings):
            env_name = match.group('name')
            has_project_id = match.group('first_separator') == ':'
            has_default = match.group('second_separator') == ':'

            if not has_project_id:
                raise ValueError(
                    f'Project id is not specified for secret {env_name}'
                )
            project_id = match.group('project_id')

            value = get_google_cloud_secret(env_name, project_id)

            if value is None:
                if not has_default:
                    raise ValueError(
                        f'Missing required google cloud secret manager '
                        f'variable {env_name}'
                    )
                value = match.group('default')
            span_min, span_max = match.span()
            config_content = (
                f'{config_content[:span_min]}'
                f'{value}'
                f'{config_content[span_max:]}'
            )
        return config_content


class GoogleLoad(Load):
    def __init__(self, loaders: Optional[List[Loader]] = None) -> None:
        super().__init__(loaders)
        self.loaders.append(GoogleLoader())



import os
from dataclasses import fields
from typing import Dict, Type, TypeVar

import yaml
from dotenv import find_dotenv

T = TypeVar('T')


def read_config(klass: Type[T], path: str = 'config/config.yaml') -> T:
    config_dict = read_yaml_to_dict(path)
    dataclass_instance = dataclass_from_dict(klass, config_dict)
    return dataclass_instance


def read_yaml_to_dict(path: str = 'config/config.yaml') -> Dict:
    # TODO: add exception for non-dict structures?
    config_abs_path = find_dotenv(path)
    if not config_abs_path:
        raise ValueError('Config yaml file not found')

    class Loader(yaml.SafeLoader):
        def __init__(self, stream):
            self._root = os.path.split(stream.name)[0]
            super().__init__(stream)

        def include(self, node):
            filename = os.path.join(self._root, self.construct_scalar(node))
            with open(filename, 'r') as file:
                return yaml.load(file, Loader)

    Loader.add_constructor('!include', Loader.include)

    with open(
        config_abs_path,
        'r',
    ) as f:
        config_dict: Dict = yaml.load(f, Loader=Loader)
    return config_dict


def dataclass_from_dict(klass: Type[T], d: Dict) -> T:
    # TODO: add exceptions or warnings for container types
    #  they work, but descriptor loaders.loaders.Load() is not working for
    #  items of list may be I should use dacite lib, idk, but it looks
    #  overcomplicated in configuration context
    #  example of config.yaml:
    #  field: "${PATH}" # Load() is working, value of env var PATH will be here
    #  lst:
    #    - key: "${secret1}" # Load() is not working
    #    - "${secret2}" # Load() is not working
    #  result will be just yaml values without loading from env:
    #  {"lst": [{"key": "secret1"}, "secret2"]}
    try:
        field_types = {field.name: field.type for field in fields(klass)}
        key_value = {}
        for key, value in d.items():
            try:
                key_value[key] = dataclass_from_dict(field_types[key], value)
            except KeyError as _:
                key_value[key] = value
        return klass(**key_value)
    except TypeError as e:
        if 'must be called with a dataclass type or instance' in str(e):
            return d
        raise e



from dataclasses import dataclass
from typing import List

from dyacon import read_config
from dyacon.read import read_yaml_to_dict, dataclass_from_dict
import pytest


def test_read_yaml_to_dict():
    assert read_yaml_to_dict("tests/config/config0.yaml") == {
        "db": {"name": "db_name", "port": 1234}
    }


def test_read_yaml_to_dict_raise():
    with pytest.raises(ValueError, match="Config yaml file not found"):
        read_yaml_to_dict("non_existed_file.yaml")


def test_include():
    assert read_yaml_to_dict("tests/config/config1.yaml") == {
        "db": {"name": "db_name", "port": 1234}
    }


def test_dataclass_from_dict():
    @dataclass
    class SubClass:
        key1: str
        key2: int

    @dataclass
    class Config:
        key1: str
        key2: int
        key3: List[str]
        key4: SubClass
        key5: List[SubClass]

    assert dataclass_from_dict(
        Config,
        {
            "key1": "value1",
            "key2": 2,
            "key3": ["value3", "value4"],
            "key4": {
                "key1": "value1",
                "key2": 2,
            },
            "key5": [
                {
                    "key1": "value1",
                    "key2": 2,
                },
            ],
        },
    ) == Config(
        key1="value1",
        key2=2,
        key3=["value3", "value4"],
        key4=SubClass(key1="value1", key2=2),
        key5=[{"key1": "value1", "key2": 2}],  # noqa
        # IMPORTANT: key5 is not SubClass because it is a list of instances
        # and dataclass_from_dict can't handle it
    )


def test_dataclass_from_dict_raise():
    @dataclass
    class Config:
        key1: str
        key2: int

    with pytest.raises(TypeError, match="got an unexpected keyword argument"):
        dataclass_from_dict(
            Config,
            {
                "key1": "value1",
                "key2": 2,
                "key3": ["value3", "value4"],
            },
        )

    with pytest.raises(
        TypeError,
        match="missing 1 required positional argument",
    ):
        dataclass_from_dict(
            Config,
            {
                "key1": "value1",
            },
        )


def test_read_config():
    @dataclass
    class DB:
        name: str
        port: int

    @dataclass
    class Config:
        db: DB

    assert read_config(Config, "tests/config/config0.yaml") == Config(
        db=DB(
            name="db_name",
            port=1234,
        )
    )



from dataclasses import dataclass
from unittest.mock import patch

import pytest
from dyacon import Load
from dyacon.read import dataclass_from_dict
from dyacon.loaders.external import GoogleLoad


def test_env_loader():
    @dataclass
    class Config:
        key1: str = Load()

    config_dict = {'key1': '${ENV1}'}

    with patch('os.getenv', return_value='value1'):
        assert dataclass_from_dict(Config, config_dict) == Config('value1')


def test_env_loader_raise():
    @dataclass
    class Config:
        key1: str = Load()

    config_dict = {'key1': '${ENV1}'}

    with patch('os.getenv', return_value=None):
        with pytest.raises(
            ValueError,
            match='Missing required environment variable ENV1',
        ):
            dataclass_from_dict(Config, config_dict)


def test_env_loader_default_value():
    @dataclass
    class Config:
        key1: str = Load()

    config_dict = {'key1': '${ENV1:default}'}

    with patch('os.getenv', return_value=None):
        assert dataclass_from_dict(Config, config_dict) == Config('default')


def test_env_loader_several_groups():
    @dataclass
    class Config:
        key1: str = Load()

    config_dict = {'key1': 'hello-${ENV1}${ENV2}'}

    with patch.dict('os.environ', {'ENV1': 'value1', 'ENV2': 'value2'}):
        assert dataclass_from_dict(Config, config_dict) == Config(
            'hello-value1value2'
        )


def test_google_loader_default():
    @dataclass
    class Config:
        key1: str = GoogleLoad()

    config_dict = {'key1': '!{secret_name:project_id:default}'}

    assert dataclass_from_dict(Config, config_dict) == Config('default')


def test_google_loader_without_default():
    @dataclass
    class Config:
        key1: str = GoogleLoad()

    config_dict = {'key1': '!{secret_name:project_id}'}

    with pytest.raises(
        ValueError,
        match='Missing required google cloud secret manager',
    ):
        dataclass_from_dict(Config, config_dict)



import os
import re
from abc import ABC, abstractmethod
from typing import List, Optional, Pattern


class Loader(ABC):
    @property
    @abstractmethod
    def pattern(self) -> Pattern[str]:
        raise NotImplementedError

    @abstractmethod
    def load(self, config_content: str) -> str:
        raise NotImplementedError


ENV_LOADER_PATTERN = re.compile(
    r'\${'
    r'(?P<name>[^}^{:]+)'
    r'(?P<separator>:?)'
    r'(?P<default>.*?)'
    r'}'
)


class EnvLoader(Loader):
    def __init__(self) -> None:
        self._pattern = ENV_LOADER_PATTERN

    @property
    def pattern(self) -> Pattern[str]:
        return self._pattern

    def load(self, config_content: str) -> str:
        findings = list(self.pattern.finditer(config_content))

        for match in reversed(findings):
            env_name = match.group('name')
            has_default = match.group('separator') == ':'

            value = os.getenv(env_name)

            if value is None:
                if not has_default:
                    raise ValueError(
                        f'Missing required environment variable {env_name}'
                    )
                value = match.group('default')
            span_min, span_max = match.span()
            config_content = (
                f'{config_content[:span_min]}'
                f'{value}'
                f'{config_content[span_max:]}'
            )
        return config_content


class Load:
    def __init__(self, loaders: Optional[List[Loader]] = None) -> None:
        self.loaders = [EnvLoader()]
        if loaders:
            self.loaders.extend(loaders)

    def load(self, config_content: str) -> str:
        for loader in self.loaders:
            if list(loader.pattern.finditer(config_content)):
                return loader.load(config_content)
        return config_content

    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, instance, owner=None):
        if instance is None:
            return self
        return instance.__dict__[self.name]

    def __set__(self, instance, value):
        instance.__dict__[self.name] = self.load(value)
